/***************************************************************************
 *
 * gpm.c : GNOME Power Manager
 *
 * Copyright (C) 2005 Richard Hughes, <richard@hughsie.com>
 *
 * Taken in part from:
 * - lshal   (C) 2003 David Zeuthen, <david@fubar.dk>
 * - notibat (C) 2004 Benjamin Kahn, <xkahn@zoned.net>
 *
 * Licensed under the GNU GPL v2.  See COPYING.
 *
 **************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <glib.h>
#include <dbus/dbus-glib.h>

#include <sys/file.h>
#include <X11/X.h>

#include <gdk/gdkx.h>
#include <gnome.h>
#include <glade/glade.h>
#include <gconf/gconf-client.h>

#include <libhal.h>
#include "common.h"
#include "eggtrayicon.h"
#include "gpm.h"
#include "logger.h"
#include "object_table.h"
#include "clipboard.h"

static LibHalContext *hal_ctx;
static HasData has_data;
static MenuData *menu_data;
static StateData state_data;
static gboolean system_bus, is_verbose;

/*static gboolean hasIconSemophore; TODO, to avoid races */
GenericObject objectData[MAX_ICONS];


/** Do an interactive alert, i.e. spawn g-p-a
 *
 *  @param  text		the text to be used in the dialogue, e.g. "Logitech MX-1000 mouse"
 *  @param  signal		the signal to use.
 */
static void
do_interactive_alert (const char *text, const char *signal)
{
	GPM_TRACE (("do_interactive_alert"));
	gboolean retval;
	char *path;
	setenv("GPM_ALERT_TEXT", text, 1);
	setenv("GPM_ALERT_SIGNAL", signal, 1);

	path = g_strconcat (BINDIR, "/", "gnome-power-alert", NULL);
	retval = g_spawn_command_line_async (path, NULL);
	if (retval == FALSE)
		GPM_ERROR (("Couldn't execute command: %s", path));
	g_free (path);
}

/** Gets policy from gconf
 *
 *  @param  name		gconf policy name
 *  @return 			the int gconf value of the policy
 */
static int
getPolicy (const char *name)
{
	gchar *gconfpath;
	int value;
	GConfClient *client = gconf_client_get_default ();
	gconfpath = g_strconcat ("/apps/gnome-power/policy/", name, NULL);
	value = gconf_client_get_int (client, gconfpath, NULL);
	g_free (gconfpath);
	return value;
}

/** Do the action dictated by policy
 *
 *  @param  name		gconf policy name
 */
static void
actionPolicyDo (const char *name)
{
	int policy_number;
	policy_number = getPolicy (name);
	GPM_INFO (("%s policy is '%i'", name, policy_number));
	if (policy_number == ACTION_NOTHING)
		GPM_INFO (("*ACTION* Doing nothing"));
	else if (policy_number == ACTION_WARNING) {
		GPM_INFO (("*ACTION* Send warning"));
		/* testing only */
		do_interactive_alert ("*Testing* Logitech MX-1000", "ups_low");
	} else if (policy_number == ACTION_SUSPEND)
		GPM_INFO (("*ACTION* Suspend"));
	else if (policy_number == ACTION_HIBERNATE)
		GPM_INFO (("*ACTION* Hibernate"));
	else if (policy_number == ACTION_SHUTDOWN)
		GPM_INFO (("*ACTION* Shutdown"));
}

/** Do the percentage dictated by policy
 *
 *  @param  xxxx		test
 */
static void
actionPolicyPercent (const char *name)
{
	int value;
	value = getPolicy (name);
	GPM_INFO (("*ACTION* LCD Brightness to %i%%", value));
}

/** Compare the old and the new values, if different or force'd then updates gconf
 *
 *  @param  name		general gconf name, e.g. hasButtons
 */
static void
compare_bool_set_gconf (const char *name, gboolean *has_dataold, gboolean *has_datanew, gboolean force)
{
	GPM_TRACE (("compare_bool_set_gconf"));
	gchar *gconfpath = g_strconcat ("/apps/gnome-power/general/", name, NULL);
	GConfClient *client = gconf_client_get_default ();

	if (force || *has_datanew != *has_dataold) {
		*has_dataold = *has_datanew;
		gconf_client_set_bool (client, gconfpath, *has_datanew, NULL);
		GPM_INFO (("%s = %i", name, *has_datanew));
	}
	g_free (gconfpath);
}

/** Recalculate logic of HasData, without any DBUS, all cached internally
 *  All complicated convoluted logic goes here :-)
 *  @param  coldplug		If set, send events even if they are the same
 */
void
updateHasLogic (GenericObject *slotBase, gboolean coldplug)
{
	GPM_TRACE (("updateHasLogic"));
	if (!slotBase) {
		GPM_ERROR (("updateHasLogic : slotBase is NULL"));
		return;
	}
		
	int a;
	/* set up temp. state */
	HasData has_datanew;
	has_datanew.hasBatteries = FALSE;
	has_datanew.hasAcAdapter = FALSE;
	has_datanew.hasButtons = TRUE;
	has_datanew.hasUPS = FALSE;
	has_datanew.hasDisplays = FALSE;
	has_datanew.hasHardDrive = FALSE;
	has_datanew.hasLCD = FALSE;
	has_datanew.hasWireless = FALSE;

	for (a=0; a<MAX_ICONS; a++) {
		if (slotBase->powerDevice == POWER_UPS)
			has_datanew.hasUPS = TRUE;
		else if (slotBase->powerDevice == POWER_MOUSE)
			has_datanew.hasWireless = TRUE;
		else if (slotBase->powerDevice == POWER_KEYBOARD)
			has_datanew.hasWireless = TRUE;
		else if (slotBase->powerDevice == POWER_BATTERY)
			has_datanew.hasBatteries = TRUE;
		else if (slotBase->powerDevice == POWER_AC_ADAPTER)
			has_datanew.hasAcAdapter = TRUE;
		slotBase++;
	}
	compare_bool_set_gconf ("hasUPS", &has_data.hasUPS, &has_datanew.hasUPS, coldplug);
	compare_bool_set_gconf ("hasButtons", &has_data.hasButtons, &has_datanew.hasButtons, coldplug);
	compare_bool_set_gconf ("hasAcAdapter", &has_data.hasAcAdapter, &has_datanew.hasAcAdapter, coldplug);
	compare_bool_set_gconf ("hasBatteries", &has_data.hasBatteries, &has_datanew.hasBatteries, coldplug);
	compare_bool_set_gconf ("hasWireless", &has_data.hasWireless, &has_datanew.hasWireless, coldplug);
	compare_bool_set_gconf ("hasLCD", &has_data.hasLCD, &has_datanew.hasLCD, coldplug);
	compare_bool_set_gconf ("hasHardDrive", &has_data.hasHardDrive, &has_datanew.hasHardDrive, coldplug);
	compare_bool_set_gconf ("hasDisplays", &has_data.hasDisplays, &has_datanew.hasDisplays, coldplug);

	/* update state data as this may have changed */
}

/** Description
 *
 *  @param  xxxx		test
 */
static void
mememe (int action)
{
	if (action == ACTION_POWER_STATE_CHANGE) {
		GPM_INFO (("DBUS: %s = %i", "onBatteryPower", state_data.onBatteryPower));
		if (has_data.hasDisplays && has_data.hasLCD) {
			if (state_data.onBatteryPower)
				actionPolicyPercent ("LCDBatteryPercentage");
			else
				actionPolicyPercent ("LCDMainsPercentage");
		}
		if (has_data.hasBatteries) {
			if (state_data.onBatteryPower)
				actionPolicyDo ("BatteryFailAction");
		}
	}
}

/** Recalculate logic of StateData, without any DBUS, all cached internally
 * Exported DBUS interface values goes here :-)
 *  @param  coldplug		If set, send events even if they are the same
 */
void
updateStateLogic (GenericObject *slotBase, gboolean coldplug)
{
	GPM_TRACE (("updateStateLogic"));
	if (!slotBase) {
		GPM_ERROR (("updateStateLogic : slotBase is NULL"));
		return;
	}
	int a;
	GenericObject *slotBaseTemp;
	slotBaseTemp = slotBase;
	/* set up temp. state */
	StateData state_datanew;
	state_datanew.onBatteryPower = TRUE;
	state_datanew.onUPSPower = FALSE;

	for (a=0; a<MAX_ICONS; a++) {
		if (slotBase->powerDevice == POWER_UPS && !slotBase->isCharging)
			state_datanew.onUPSPower = TRUE;
		slotBase++;
	}

	/* get old value */
	slotBase = slotBaseTemp;
	if (has_data.hasBatteries == TRUE) {
		/* Reverse logic as only one ac_adapter is needed to be "on mains power" */
		for (a=0; a<MAX_ICONS; a++) {
			if (slotBase->powerDevice==POWER_AC_ADAPTER) {
				state_datanew.onBatteryPower = FALSE;
				break;
				}
			slotBase++;
		}
	}

	if (coldplug || state_datanew.onBatteryPower != state_data.onBatteryPower) {
		state_data.onBatteryPower = state_datanew.onBatteryPower;
		mememe (ACTION_POWER_STATE_CHANGE);
		/*GPM_INFO (("DBUS: %s = %i", "onBatteryPower", state_data.onBatteryPower));
		if (has_data.hasDisplays) {
			if (state_data.onBatteryPower)
				actionPolicyPercent ("lcd_batteries");
			else
				actionPolicyPercent ("lcd_mains");
		}
		if (has_data.hasBatteries) {
			if (state_data.onBatteryPower)
				actionPolicyDo ("combo_battery_fail");
		}*/
	}

	if (coldplug || state_datanew.onUPSPower != state_data.onUPSPower) {
		state_data.onUPSPower = state_datanew.onUPSPower;
		GPM_INFO (("DBUS: %s = %i", "onUPSPower", state_data.onUPSPower));
	}

}

/** Convenience function.
 *  Prints errors due to wrong values expected (exposes bugs, rather than hides them)
 */
static void
dbus_error_print (DBusError *error)
{
	if (dbus_error_is_set (error)) {
		GPM_ERROR (("Oops: Internal DBUS Error:%s", error->message));
	}
}

/** Create new powerdevice icon on the panel.
 *
 *  @param  slotData		the GenericObject reference
 *  @param  menu_data			MenuData of the tray
 */
static void
icon_add_powerdevice (GenericObject *slotData, MenuData *menu_data)
{
	GPM_TRACE (("icon_add_powerdevice"));
	if (!slotData) {
		GPM_ERROR (("icon_add_powerdevice : slotData is NULL"));
		return;
	}
	char filename[128];
	char tooltip[128];
	gboolean show;

	if (slotData->hasIcon) {
		GPM_INFO (("Not creating icon '%i', as already been set!", slotData->powerDevice));
		return;
		}

	GConfClient *client = gconf_client_get_default ();
	show = FALSE;
	if (slotData->powerDevice == POWER_MOUSE)
		show = gconf_client_get_bool (client, "/apps/gnome-power/icons/showWireless", NULL);
	else if (slotData->powerDevice == POWER_KEYBOARD)
		show = gconf_client_get_bool (client, "/apps/gnome-power/icons/showWireless", NULL);
	else if (slotData->powerDevice == POWER_BATTERY)
		show = gconf_client_get_bool (client, "/apps/gnome-power/icons/showBatteries", NULL);
	else if (slotData->powerDevice == POWER_AC_ADAPTER)
		show = gconf_client_get_bool (client, "/apps/gnome-power/icons/showAcAdapter", NULL);
	else if (slotData->powerDevice == POWER_UPS)
		show = gconf_client_get_bool (client, "/apps/gnome-power/icons/showUPS", NULL);
	else
		GPM_WARNING (("icon_add_powerdevice was asked to add an unknown device"));

	if (!show) {
		GPM_INFO (("Not creating icon '%i', as gconf says NO!", slotData->powerDevice));
		return;
	} else
		GPM_INFO (("Creating icon '%i'", slotData->powerDevice));

	updatePercentageChargeFromSlot (slotData);
	updatePowerStateFromSlot (slotData);
	updateMinutesRemainingFromSlot (slotData);

	getTooltipFromSlot (slotData, tooltip);
	getIconFilenameFromSlot (slotData, filename);

	slotData->hasIcon = TRUE;
	icon_create (menu_data, slotData->slot, filename, tooltip);
}

/** Description
 *
 *  @param  xxxx		test
 */
static void
icon_remove_powerdevice (GenericObject *slotData, MenuData *menu_data)
{
	GPM_TRACE (("icon_remove_powerdevice"));
	if (!slotData) {
		GPM_ERROR (("icon_remove_powerdevice : slotData is NULL"));
		return;
	}
	if (slotData->hasIcon) {
		slotData->hasIcon = FALSE;
		icon_destroy (menu_data->td[slotData->slot]);
	} else
		GPM_INFO (("Not removing icon '%i', as already been removed!", slotData->powerDevice));
}

/** Generic exit
 *
 */
static void
gpm_exit (void)
{
	int a;
	GPM_INFO (("Shutting down..."));

	/* deallocates menu and all icons */
	for (a=0; a<MAX_ICONS; a++)
		icon_destroy (menu_data->td[a]);
	g_free(menu_data);

	exit (0);
}

/** Callback for "about" box
 *
 */
static void
callback_about_activated (GtkMenuItem *menuitem, gpointer user_data)
{
	GPM_TRACE (("callback_about_activated"));
	static GtkWidget *about = NULL;
	GdkPixbuf *pixbuf = NULL;
	const gchar *authors[] = { "Richard Hughes <richard@hughsie.com>", NULL };
	const gchar *documenters[] = { NULL };
	const gchar *translator_credits = _("translator_credits");

	if (about != NULL)
	{
		gdk_window_raise (about->window);
		gdk_window_show (about->window);
		return;
	}

	pixbuf = gdk_pixbuf_new_from_file (GPM_DATA "battery-48.png", NULL);
	about = gnome_about_new(NICENAME, VERSION,
			"Copyright \xc2\xa9 2005 Richard Hughes",
			_(NICEDESC),
			(const char **)authors,
			(const char **)documenters,
			strcmp (translator_credits, "translator_credits") != 0 ? translator_credits : NULL,
			pixbuf);

	if (pixbuf != NULL)
		gdk_pixbuf_unref (pixbuf);

	g_signal_connect (G_OBJECT (about), "destroy", G_CALLBACK (gtk_widget_destroyed), &about);
	g_object_add_weak_pointer (G_OBJECT (about), (void**)&(about));
	gtk_widget_show(about);
}

/** Callback for quit
 *
 */
static void
callback_quit_activated (GtkMenuItem *menuitem, gpointer user_date)
{
	GPM_TRACE (("callback_quit_activated"));
	GnomeClient *master;
	GnomeClientFlags flags;
	master = gnome_master_client ();
	flags = gnome_client_get_flags (master);
	if (flags & GNOME_CLIENT_IS_CONNECTED) {
		gnome_client_set_restart_style (master,	GNOME_RESTART_NEVER);
		gnome_client_flush (master);
	}
	gpm_exit ();
}

/** Callback for preferences
 *
 */
static void
callback_prefs_activated (GtkMenuItem *menuitem, gpointer user_data)
{
	GPM_TRACE (("callback_prefs_activated"));
	gboolean retval;
	char *path;

	path = g_strconcat (BINDIR, "/", "gnome-power-preferences", NULL);
	retval = g_spawn_command_line_async (path, NULL);
	if (retval == FALSE)
		GPM_ERROR (("Couldn't execute command: %s", path));
	g_free (path);
}

/** Creates panel menu
 *
 *  @return			MenuData object
 */
static MenuData *
menu_create (void)
{
	GPM_TRACE (("menu_create"));
	GtkWidget *item;
	MenuData *menu_data;
	int a;

	/* Popup menu */
	menu_data = g_new0 (MenuData, 1);
	menu_data->popup_menu = gtk_menu_new ();

	/* Initialise to known state */
	for (a=0; a<MAX_ICONS; a++)
		menu_data->td[a] = NULL;

	item = gtk_image_menu_item_new_from_stock (GTK_STOCK_PREFERENCES, NULL);
	g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (callback_prefs_activated), (gpointer) menu_data);
	gtk_widget_show (item);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu_data->popup_menu), item);

	item = gtk_image_menu_item_new_from_stock (GNOME_STOCK_ABOUT, NULL);
	g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (callback_about_activated), (gpointer) menu_data);
	gtk_widget_show (item);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu_data->popup_menu), item);

	item = gtk_image_menu_item_new_from_stock (GTK_STOCK_QUIT, NULL);
	g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK(callback_quit_activated), (gpointer) menu_data);
	gtk_widget_show(item);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu_data->popup_menu), item);
	return menu_data;
}

/** Initialises GNOME program.
 *
 */
static void
init_gnome (void)
{
	GPM_TRACE (("init_gnome"));
	GnomeClient *master;
	GnomeClientFlags flags;
	master = gnome_master_client ();
	flags = gnome_client_get_flags (master);
	if (flags & GNOME_CLIENT_IS_CONNECTED) {
		gnome_client_set_restart_style (master, GNOME_RESTART_IMMEDIATELY);
		gnome_client_flush (master);
	}
	g_signal_connect (GTK_OBJECT (master), "die", G_CALLBACK (gpm_exit), NULL);
}

/** Adds a ac_adapter device. Also sets up properties on cached object
 *
 *  @param  udi			UDI
 */
static void
add_ac_adapter (const char *udi)
{
	GPM_TRACE (("add_ac_adapter"));
	DBusError error;
	GenericObject *slotData;
	dbus_error_init (&error);
	if (libhal_device_get_property_bool (hal_ctx, udi, "ac_adapter.present", &error)) {
		if (object_table_add (udi, &slotData, objectData)) {
			slotData->powerDevice = POWER_AC_ADAPTER;
			icon_add_powerdevice(slotData, menu_data);
			GPM_INFO (("ac_adapter added"));
		} else
			GPM_WARNING (("ac_adapter already added"));
	} else {
		GPM_INFO (("ac adapter device ignored as not present : %s", udi));
	}
}

/** Adds a battery device, of any type. Also sets up properties on cached object
 *
 *  @param  udi			UDI
 */
static void
add_battery (const char *udi)
{
	GPM_TRACE (("add_battery"));
	int powerDevice;
	DBusError error;
	char *type, *tech;
	GenericObject *slotData;

	dbus_error_init (&error);
	type = libhal_device_get_property_string (hal_ctx, udi, "battery.type", &error);
	dbus_error_print (&error);

	/* battery is further refined using the .type property */
	if (strcmp (type, "ups") == 0)
		powerDevice = POWER_UPS;
	else if (strcmp (type, "mouse") == 0)
		powerDevice = POWER_MOUSE;
	else if (strcmp (type, "keyboard") == 0)
		powerDevice = POWER_KEYBOARD;
	else
		powerDevice = POWER_BATTERY;

	dbus_error_init (&error);
	if (powerDevice==POWER_UPS || libhal_device_get_property_bool (hal_ctx, udi, "battery.present", &error)) {
		if (object_table_add (udi, &slotData, objectData)) {
			if (powerDevice==POWER_BATTERY)
				GPM_INFO (("battery added"));
			else if (powerDevice==POWER_UPS)
				GPM_INFO (("ups added"));
			else if (powerDevice==POWER_MOUSE)
				GPM_INFO (("battery mouse added"));
			else if (powerDevice==POWER_KEYBOARD)
				GPM_INFO (("battery keyboard added"));

			/* set cached variables up */
			dbus_error_init (&error);
			tech = libhal_device_get_property_string (hal_ctx, udi, "battery.charge_level.unit", &error);

			
			if (tech && strcmp (tech, "percent") == 0) {
				/* battery only gives percentage readings */
				slotData->reportType = REPORT_PERCENT;

				dbus_error_init (&error);
				slotData->rawMinutesRemaining = libhal_device_get_property_int (hal_ctx, udi, "battery.remaining_time", &error);
				dbus_error_print (&error);

			} else if (tech && strcmp (tech, "mWh") == 0) {
				/* battery gives acpi-style readings */
				slotData->reportType = REPORT_MWH;

				dbus_error_init (&error);
				slotData->rawRate = libhal_device_get_property_int (hal_ctx, udi, "battery.charge_level.rate", &error);
				dbus_error_print (&error);

				dbus_error_init (&error);
				slotData->rawDesign = libhal_device_get_property_int (hal_ctx, udi, "battery.charge_level.design", &error);
				dbus_error_print (&error);
			} else {
				/* don't know. unknown. */
				slotData->reportType = REPORT_UNKNOWN;
			}

			/* we need the RAW readings so we keep functions modular and acpi/apm neutral */
			dbus_error_init (&error);
			slotData->rawCharge = libhal_device_get_property_int (hal_ctx, udi, "battery.charge_level.current", &error);
			dbus_error_print (&error);

			/* we need the RAW readings so we can process them later on for time remaining */
			dbus_error_init (&error);
			slotData->rawLastFull = libhal_device_get_property_int (hal_ctx, udi, "battery.charge_level.last_full", &error);
			dbus_error_print (&error);

			/* battery might not be rechargeable, have to check */
			dbus_error_init (&error);
			slotData->isRechargeable = libhal_device_get_property_bool (hal_ctx, udi, "battery.is_rechargeable", &error);
			dbus_error_print (&error);
			if (slotData->isRechargeable) {
				dbus_error_init (&error);
				slotData->isCharging = libhal_device_get_property_bool (hal_ctx, udi, "battery.rechargeable.is_charging", &error);
				dbus_error_print (&error);
			}

			if (powerDevice == POWER_BATTERY) {
				dbus_error_init (&error);
				/* might not exist */
				tech = libhal_device_get_property_string (hal_ctx, slotData->udi, "battery.technology", &error);
				if (tech)
					slotData->technologyType = tech;
				else
					slotData->technologyType = "General";
			} else if (powerDevice == POWER_MOUSE) {
				dbus_error_init (&error);
				/* might not exist */
				tech = libhal_device_get_property_string (hal_ctx, slotData->udi, "info.product", &error);
				if (tech)
					slotData->technologyType = tech;
				else
					slotData->technologyType = "General";
			} else
				slotData->technologyType = "Unknown";


			slotData->powerDevice = powerDevice;
			icon_add_powerdevice(slotData, menu_data);
		}
	} else {
		GPM_INFO (("Device ignored as not present : %s", udi));
	}
	libhal_free_string (type);
}

/** Coldplugs devices of type battery & ups at startup
 *
 */
static void
coldplug_devices (void)
{
	GPM_TRACE (("coldplug_devices"));
	int i, num_devices;
	char **device_names;
	DBusError error;

	/* devices of type battery */
	dbus_error_init (&error);
	device_names = libhal_find_device_by_capability (hal_ctx, "battery", &num_devices, &error);
	dbus_error_print (&error);
	if (device_names == NULL)
		GPM_DIE (("Couldn't obtain list of batteries"));
	for (i = 0; i < num_devices; i++) {
		add_battery (device_names[i]);
	}
	libhal_free_string_array (device_names);

	/* devices of type ac_adapter */
	dbus_error_init (&error);
	device_names = libhal_find_device_by_capability (hal_ctx, "ac_adapter", &num_devices, &error);
	dbus_error_print (&error);
	if (device_names == NULL)
		GPM_DIE (("Couldn't obtain list of ac_adapters"));
	for (i = 0; i < num_devices; i++)
		add_ac_adapter (device_names[i]);
	libhal_free_string_array (device_names);
}

/** Removes any type of device
 *
 *  @param  udi			UDI
 */
static void
remove_devices (const char *udi)
{
	GPM_TRACE (("remove_devices"));
	GenericObject *slotData;
	if (object_table_remove (udi, &slotData, objectData)) {
		if (slotData->powerDevice==POWER_AC_ADAPTER)
			GPM_INFO (("ac adapter has removed"));
		else if (slotData->powerDevice==POWER_BATTERY)
			GPM_INFO (("battery has removed"));
		else if (slotData->powerDevice==POWER_UPS)
			GPM_INFO (("ups has removed"));
		else if (slotData->powerDevice==POWER_MOUSE)
			GPM_INFO (("mouse has removed"));
		else if (slotData->powerDevice==POWER_KEYBOARD)
			GPM_INFO (("keyboard has removed"));
		else if (slotData->powerDevice==POWER_UNKNOWN)
			GPM_INFO (("unknown has removed"));
		icon_remove_powerdevice (slotData, menu_data);
		/* mark as available */
		slotData->powerDevice = POWER_NONE;
	}
}

/** Invoked when a device is removed from the Global Device List. Simply
 *  prints a message on stderr.
 *
 *  @param  udi			UDI
 */
static void
device_removed (LibHalContext *ctx, const char *udi)
{
	GPM_TRACE (("device_removed"));
	/*
	 * UPS's/mice/keyboards don't use battery.present
	 * they just disappear from the device tree
	 */
	remove_devices (udi);
	/* our state has changed, update */
	updateHasLogic (objectData, FALSE);
	updateStateLogic (objectData, FALSE);
}

/** Invoked when device in the Global Device List acquires a new capability.
 *  Prints the name of the capability to stderr.
 *
 *  @param  udi			UDI
 *  @param  capability          Name of capability
 */
static void
device_new_capability (LibHalContext *ctx, const char *udi, const char *capability)
{
	GPM_TRACE (("device_new_capability"));
	/*
	 * UPS's/mice/keyboards don't use battery.present
	 * they just appear in the device tree
	 */
	if (strcmp (capability, "battery") == 0) {
		add_battery (udi);
		/* our state has changed, update */
		updateHasLogic (objectData, FALSE);
		updateStateLogic (objectData, FALSE);
	}
}

/** Invoked when a property of a device in the Global Device List is
 *  changed, and we have we have subscribed to changes for that device.
 *
 *  @param  udi                 Univerisal Device Id
 *  @param  key                 Key of property
 */
static void
property_modified (LibHalContext *ctx, const char *udi, const char *key,
		   dbus_bool_t is_removed, dbus_bool_t is_added)
{
	GPM_TRACE (("property_modified"));
	int temp;
	DBusError error;
	int percentageCharge, modifytype, powerState;
	gboolean value;
	GenericObject *slotData;
	char filename[128];
	char tooltip[128];

	/* only process modified entries, not added or removed keys */
	if (is_removed||is_added)
		return;

	GPM_TRACE (("key = '%s'", key));

	modifytype = POWER_UNKNOWN;
	if (strcmp (key, "battery.present") == 0) {
		modifytype = POWER_BATTERY;
		dbus_error_init (&error);
		value = libhal_device_get_property_bool (ctx, udi, key, &error);
		dbus_error_print (&error);
		if (value)
			add_battery (udi);
		else
			remove_devices (udi);
		/* our state has changed, update */
		updateHasLogic (objectData, FALSE);
		updateStateLogic (objectData, FALSE);

	} else if (strcmp (key, "ac_adapter.present") == 0) {
		modifytype = POWER_AC_ADAPTER;
		dbus_error_init (&error);
		value = libhal_device_get_property_bool (ctx, udi, key, &error);
		dbus_error_print (&error);
		GPM_INFO (("value = '%i'", value));
		if (value)
			add_ac_adapter (udi);
		else
			remove_devices (udi);
		/* our state has changed, update */
		updateHasLogic (objectData, FALSE);
		updateStateLogic (objectData, FALSE);

	} else if (strcmp (key, "battery.rechargeable.is_charging") == 0) {
		dbus_error_init (&error);
		value = libhal_device_get_property_bool (ctx, udi, key, &error);
		if (object_table_find (udi, &slotData, objectData)) {
			slotData->isCharging = value;
			updateMinutesRemainingFromSlot (slotData);
			if (slotData->hasIcon) {
				getTooltipFromSlot (slotData, tooltip);
				icon_modify_tooltip (menu_data->td[slotData->slot], tooltip);
			} else
				GPM_INFO (("Not modifying icon as not yet created"));
			/* our state has changed, update */
			updateHasLogic (objectData, FALSE);
			updateStateLogic (objectData, FALSE);
		}

	} else if (strcmp (key, "battery.rechargeable.is_discharging") == 0) {
		dbus_error_init (&error);
		value = libhal_device_get_property_bool (ctx, udi, key, &error);
		dbus_error_print (&error);
		if (object_table_find (udi, &slotData, objectData)) {
			slotData->isCharging = !value;
			updateMinutesRemainingFromSlot (slotData);
			if (slotData->hasIcon) {
				getTooltipFromSlot (slotData, tooltip);
				icon_modify_tooltip (menu_data->td[slotData->slot], tooltip);
			} else
				GPM_INFO (("Not modifying icon as not yet created"));
			/* our state has changed, update */
			updateHasLogic (objectData, FALSE);
			updateStateLogic (objectData, FALSE);
		}

	} else if (strcmp (key, "battery.charge_level.current") == 0) {
		dbus_error_init (&error);
		temp = libhal_device_get_property_int (ctx, udi, key, &error);
		dbus_error_print (&error);

		/* find out whether we need to update the icon or tooltip */
		if (object_table_find (udi, &slotData, objectData)) {

			/* store raw value so we can work out time remaining in mWh */
			slotData->rawCharge = temp;

			/* save old values */
			percentageCharge = slotData->percentageCharge;
			powerState = slotData->powerState;

			/* update values */
			updatePercentageChargeFromSlot (slotData);
			updatePowerStateFromSlot (slotData);
			if (slotData->percentageCharge!=percentageCharge && slotData->hasIcon) {
				getTooltipFromSlot (slotData, tooltip);
				icon_modify_tooltip (menu_data->td[slotData->slot], tooltip);
			}
			if (slotData->powerState!=powerState && slotData->hasIcon) {
				getIconFilenameFromSlot (slotData, filename);
				icon_modify_image (menu_data->td[slotData->slot], filename);
				/* our state has changed, update */
				updateHasLogic (objectData, FALSE);
				updateStateLogic (objectData, FALSE);
			}
		} else {
			GPM_WARNING (("battery cannot be updated as not added to objectData"));
		}
	} else if (strcmp (key, "battery.charge_level.rate") == 0) {
		if (object_table_find (udi, &slotData, objectData)) {
			dbus_error_init (&error);
			temp = libhal_device_get_property_int (ctx, udi, key, &error);
			dbus_error_print (&error);
			slotData->rawRate = temp;

			/* update values */
			updateMinutesRemainingFromSlot (slotData);
			if (slotData->hasIcon) {
				getTooltipFromSlot (slotData, tooltip);
				icon_modify_tooltip (menu_data->td[slotData->slot], tooltip);
			} else
				GPM_INFO (("Not modifying icon as not yet created"));
		}
	}
}

/** Invoked when a property of a device in the Global Device List is
 *  changed, and we have we have subscribed to changes for that device.
 *
 *  @param  udi                 Univerisal Device Id
 *  @param  condition_name      Name of condition
 *  @param  message             D-BUS message with parameters
 */
static void
device_condition (LibHalContext *ctx,
		  const char *udi, 
		  const char *condition_name,
		  const char *condition_details)
{
	DBusError error;
	char *type;
	/* button.type */

	if (strcmp (condition_name, "ButtonPressed") == 0) {
		dbus_error_init (&error);
		type = libhal_device_get_property_string (hal_ctx, udi, "button.type", &error);
		dbus_error_print (&error);
		GPM_INFO (("ButtonPressed : %s", type));
		if (strcmp (type, "power") == 0) {
			actionPolicyDo ("ButtonPowerAction");
		} else if (strcmp (type, "sleep") == 0) {
			actionPolicyDo ("ButtonSleepAction");
		} else if (strcmp (type, "lid") == 0) {
			gboolean value;
			/* we only do a lid event when the lid is OPENED */
			dbus_error_init (&error);
			value = libhal_device_get_property_bool (ctx, udi, "button.state.value", &error);
			dbus_error_print (&error);
			if (value)
				actionPolicyDo ("ButtonLidAction");
		} else
			GPM_ERROR (("Button '%s' unrecognised", type));
	}
}

/** LibHal setup
 *
 *  @param  conn		D-BUS connection
 *  @return			D-BUS connection or NULL on error
 */
static DBusConnection *
libhal_setup ()
{
	GPM_TRACE (("libhal_setup"));

	DBusConnection *conn = NULL;
	DBusError error;
	dbus_error_init (&error);
	conn = dbus_bus_get (DBUS_BUS_SYSTEM, &error);
	if (conn == NULL) {
		fprintf (stderr, "error: dbus_bus_get: %s: %s\n", error.name, error.message);
		return NULL;
	}
	if ((hal_ctx = libhal_ctx_new ()) == NULL) {
		fprintf (stderr, "error: libhal_ctx_new\n");
		return NULL;
	}
	if (!libhal_ctx_set_dbus_connection (hal_ctx, conn)) {
		fprintf (stderr, "error: libhal_ctx_set_dbus_connection: %s: %s\n", error.name, error.message);
		return NULL;
	}
	if (!libhal_ctx_init (hal_ctx, &error)) {
		fprintf (stderr, "error: libhal_ctx_init: %s: %s\n", error.name, error.message);
		return NULL;
	}
	libhal_ctx_set_device_property_modified (hal_ctx, property_modified);
	libhal_ctx_set_device_removed (hal_ctx, device_removed);
	libhal_ctx_set_device_new_capability (hal_ctx, device_new_capability);
	libhal_ctx_set_device_condition (hal_ctx, device_condition);
	return conn;
}

/** LibHal cleanup
 *
 *  @param  conn		D-BUS connection
 */
static void
libhal_cleanup (DBusConnection *conn)
{
	GPM_TRACE (("libhal_cleanup"));
	DBusError error;
	dbus_error_init (&error);
	libhal_ctx_shutdown (hal_ctx, &error);
	libhal_ctx_free (hal_ctx);
	dbus_connection_disconnect (conn);
	dbus_connection_unref (conn);
}

static void
icon_hide_or_display (const int deviceType, const gboolean value)
{
	GPM_TRACE (("icon_hide_or_display"));
	int a;
	for (a=0; a<MAX_ICONS; a++)
		if (objectData[a].powerDevice == deviceType) {
			if (value && !objectData[a].hasIcon)
				icon_add_powerdevice (&objectData[a], menu_data);
			else if (!value && objectData[a].hasIcon)
				icon_remove_powerdevice (&objectData[a], menu_data);
		}
}


static void
gconf_key_action (const char *key)
{
	GPM_TRACE (("gconf_key_action"));
	gboolean value;
	GConfClient *client = gconf_client_get_default ();

	value = gconf_client_get_bool (client, key, NULL);
	if (strcmp (key, "/apps/gnome-power/icons/showWireless") == 0) {
		icon_hide_or_display (POWER_MOUSE, value);
		icon_hide_or_display (POWER_KEYBOARD, value);
	} else if (strcmp (key, "/apps/gnome-power/icons/showUPS") == 0) {
		icon_hide_or_display (POWER_UPS, value);
	} else if (strcmp (key, "/apps/gnome-power/icons/showLCD") == 0) {
		icon_hide_or_display (POWER_LCD, value);
	} else if (strcmp (key, "/apps/gnome-power/icons/showBatteries") == 0) {
		icon_hide_or_display (POWER_BATTERY, value);
	} else if (strcmp (key, "/apps/gnome-power/icons/showAcAdapter") == 0) {
		icon_hide_or_display (POWER_AC_ADAPTER, value);
	}
}

static void
callback_gconf_key_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
	GPM_TRACE (("callback_gconf_key_changed"));
	if (gconf_entry_get_value (entry) == NULL)
		return;

	if (gconf_entry_get_value (entry)->type == GCONF_VALUE_BOOL)
		gconf_key_action (entry->key);
}

/** Prints program usage.
 *
 */
static void print_usage (void)
{
	fprintf (stderr, "\nusage : gnome-power-manager [--system-bus] [--verbose] [--help]\n");
	fprintf (stderr,
		"\n"
		"        --system-bus            Use the gconf system bus\n"
		"        --verbose               Show extra debugging\n"
		"        --help                  Show this information and exit\n"
		"\n");
}

/** Entry point
 *
 *  @param  argc		Number of arguments given to program
 *  @param  argv		Arguments given to program
 *  @return			Return code
 */
int
main (int argc, char *argv[])
{
	int a;
	GMainLoop *loop;
	DBusConnection *conn = NULL;
	DBusError error;
	dbus_error_init (&error);

	gconf_init (argc, argv, NULL);
	GConfClient *client = gconf_client_get_default ();
	gconf_client_add_dir (client, "/apps/gnome-power", GCONF_CLIENT_PRELOAD_NONE, NULL);
	gconf_client_notify_add (client, "/apps/gnome-power/icons", callback_gconf_key_changed, NULL, NULL, NULL);

	bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);

	system_bus = FALSE;
	is_verbose = FALSE;
	for (a=1; a < argc; a++) {
		if (strcmp (argv[a], "--verbose") == 0)
			is_verbose = TRUE;
		else if (strcmp (argv[a], "--system-bus") == 0)
			system_bus = TRUE;
		else if (strcmp (argv[a], "--help") == 0) {
			print_usage ();
			return EXIT_SUCCESS;
		}
	}

	if (is_verbose)
		logger_enable ();
	else
		logger_disable ();

	gnome_program_init ("GNOME Power Manager", VERSION, LIBGNOMEUI_MODULE, argc, argv, NULL);
	init_gnome ();

	if (gpm_get_lock () == FALSE)
	{
		g_print ("Already running, exiting...\n");
		exit (0);
	}

	g_print ("%s %s - %s\n", NICENAME, VERSION, NICEDESC);
	g_print ("Please report bugs to richard@hughsie.com\n");

	g_print ("\n**** WARNING ****\n");
	g_print ("Please use hal-cvs20050404 (or later) if you want APM batteries, UPS's and Wireless mice and keyboards to work. This is due to bugs only fixed recently in HAL.\n");
	g_print ("APM and PMU batteries only work at coldplug and their values do not auto-update.\n");
	g_print ("Functionality in HAL will be added soon.\n");
	g_print ("**** WARNING ****\n\n");

	loop = g_main_loop_new (NULL, FALSE);

	conn = libhal_setup ();
	if (conn == NULL)
		exit (1);

	/* get session preferences for icons */
	gconf_key_action ("/apps/gnome-power/icons/showAcAdapter");
	gconf_key_action ("/apps/gnome-power/icons/showBatteries");
	gconf_key_action ("/apps/gnome-power/icons/showUPS");
	gconf_key_action ("/apps/gnome-power/icons/showLCD");
	gconf_key_action ("/apps/gnome-power/icons/showWireless");

	dbus_connection_setup_with_g_main (conn, NULL);
	menu_data = menu_create ();

	/* initialise devices at startup */
	for (a=0; a<MAX_ICONS; a++) {
		objectData[a].powerDevice = POWER_NONE;
		objectData[a].hasIcon = FALSE;
	}

	coldplug_devices ();
	if (is_verbose)
		object_table_print (objectData);
	updateHasLogic (objectData, TRUE);
	updateStateLogic (objectData, TRUE);

	libhal_device_property_watch_all (hal_ctx, &error);

	g_main_loop_run (loop);

	libhal_cleanup (conn);

	gpm_exit();
	return 0;
}
